#include "sal/config.h"
#include "cppuhelper/factory.hxx"
#include "cppuhelper/implementationentry.hxx"
#include "cppuhelper/implbase2.hxx"
#include "com/sun/star/lang/XServiceInfo.hpp"
#include "com/sun/star/smarttags/XSmartTagRecognizer.hpp"
#include "com/sun/star/smarttags/SmartTagRecognizerMode.hpp"
#include "com/sun/star/text/XTextMarkup.hpp"
#include "com/sun/star/text/TextMarkupType.hpp"
#include "com/sun/star/i18n/XBreakIterator.hpp"
#include "com/sun/star/i18n/WordType.hpp"
#include "com/sun/star/util/SearchOptions.hpp"
#include "com/sun/star/util/SearchFlags.hpp"

#include <libxml/xmlmemory.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>
#include <unotools/textsearch.hxx>
#include <unotools/bootstrap.hxx>
#include <osl/file.hxx>
#include <tools/urlobj.hxx>
#include <tools/stream.hxx>
#include <vector>
#include <hash_set>
#include <string>

#define C2U(cChar) rtl::OUString::createFromAscii(cChar)

void lcl_GetFileList( rtl::OUString aDirectory, std::vector<rtl::OUString>& rFileList )
{
    // iterate through directory contents
    osl::FileStatus aFileStatus(FileStatusMask_All);
    osl::DirectoryItem aDirItem;
    osl::Directory aDir( aDirectory + C2U("/smarttags"));
    aDir.open();
    while ( osl::FileBase::E_None == aDir.getNextItem( aDirItem ) )
    {
        aDirItem.getFileStatus(aFileStatus);
        if ( osl::FileStatus::Regular == aFileStatus.getFileType() )
        {
            rtl::OUString ustrSystemPath;
            osl::FileBase::getSystemPathFromFileURL( aFileStatus.getFileURL(), ustrSystemPath );
            rFileList.push_back(ustrSystemPath);
        }
    }
    aDir.close();
}

/** Class: MOSTLRecognizerData

    This class is used to store all the information retrieved
    from a xml smarttag file. One object for each smart tag type.
*/
class MOSTLRecognizerData
{
private:

    rtl::OUString aCaption;
    rtl::OUString aSmartTagType;
    std::hash_set<rtl::OUString, rtl::OUStringHash> aTermList;
    std::vector<rtl::OUString> aRegExpList;

public:
    inline rtl::OUString getSmartTagType() const { return aSmartTagType; }
    inline void setSmartTagType(rtl::OUString aNewName) { this->aSmartTagType = aNewName; }

    inline rtl::OUString getCaption() const { return aCaption; }
    inline void setCaption(rtl::OUString aNewCaption) { this->aCaption = aNewCaption; }

    // term list
    inline bool checkTermList( rtl::OUString aToCheck ) const;
    inline void appendTermList(rtl::OUString aStr) { aTermList.insert(aStr); }

    // regexp list
    inline bool checkRegExps( rtl::OUString aToCheck ) const;
    inline void appendRegexList(rtl::OUString aStr) { aRegExpList.push_back(aStr); }
};

bool MOSTLRecognizerData::checkTermList( rtl::OUString aToCheck ) const
{
    // find stuff in hash_list
    return aTermList.end() != aTermList.find( aToCheck );
}

bool MOSTLRecognizerData::checkRegExps( rtl::OUString aToCheck ) const
{
    com::sun::star::util::SearchOptions aOptions;
    aOptions.algorithmType = com::sun::star::util::SearchAlgorithms_REGEXP;

    for (sal_uInt32 j=0; j < aRegExpList.size(); ++j)
    {
        const rtl::OUString aRegex = aRegExpList[j];
        aOptions.searchString = aRegex;
        utl::TextSearch aSearcher( aOptions );

        USHORT nStart = 0;
        USHORT nEnd = (USHORT)aToCheck.getLength();
        bool bFound = (aSearcher.SearchFrwrd(aToCheck, &nStart, &nEnd )!=0);
        if ( bFound )
            return true;
    }

    return false;
}

// component helper namespace
namespace comp_MOSTLSmartTagRecognizer {

namespace css = ::com::sun::star;

// component and service helper functions:
::rtl::OUString SAL_CALL _getImplementationName();
css::uno::Sequence< ::rtl::OUString > SAL_CALL _getSupportedServiceNames();
css::uno::Reference< css::uno::XInterface > SAL_CALL _create( css::uno::Reference< css::uno::XComponentContext > const & context );

} // closing component helper namespace

/// anonymous implementation namespace
namespace {

namespace css = ::com::sun::star;

class MOSTLSmartTagRecognizer:
    public ::cppu::WeakImplHelper2<
        css::lang::XServiceInfo,
        css::smarttags::XSmartTagRecognizer>
{
public:
    explicit MOSTLSmartTagRecognizer(css::uno::Reference< css::uno::XComponentContext > const & context);

    // ::com::sun::star::lang::XServiceInfo:
    virtual ::rtl::OUString SAL_CALL getImplementationName() throw (css::uno::RuntimeException);
    virtual ::sal_Bool SAL_CALL supportsService(const ::rtl::OUString & ServiceName) throw (css::uno::RuntimeException);
    virtual css::uno::Sequence< ::rtl::OUString > SAL_CALL getSupportedServiceNames() throw (css::uno::RuntimeException);

    // ::com::sun::star::lang::XInitialization:
    virtual void SAL_CALL initialize(const css::uno::Sequence< ::com::sun::star::uno::Any > & aArguments) throw (css::uno::RuntimeException, css::uno::Exception);

    // ::com::sun::star::smarttags::XSmartTagRecognizer:
    virtual ::sal_Int32 SAL_CALL getSmartTagCount() throw (css::uno::RuntimeException);
    virtual ::rtl::OUString SAL_CALL getName(const css::lang::Locale & aLocale) throw (css::uno::RuntimeException);
    virtual ::rtl::OUString SAL_CALL getDescription(const css::lang::Locale & aLocale) throw (css::uno::RuntimeException);
    virtual ::rtl::OUString SAL_CALL getSmartTagName(::sal_Int32 nSmartTagIndex) throw (css::uno::RuntimeException, css::lang::IndexOutOfBoundsException);
    virtual ::rtl::OUString SAL_CALL getSmartTagDownloadURL(::sal_Int32 nSmartTagIndex) throw (css::uno::RuntimeException, css::lang::IndexOutOfBoundsException);
    virtual void SAL_CALL recognize(const ::rtl::OUString & aText, ::sal_Int32 nStart, ::sal_Int32 nLength, css::smarttags::SmartTagRecognizerMode eDataType, const css::lang::Locale & aLocale, const css::uno::Reference< css::text::XTextMarkup > & xTextMarkup, const ::rtl::OUString & aApplicationName, const css::uno::Reference< css::frame::XController > & xController, const css::uno::Reference< css::i18n::XBreakIterator > & xTokenizer) throw (css::uno::RuntimeException);
    virtual ::sal_Bool SAL_CALL hasPropertyPage(::sal_Int32 nSmartTagIndex, const css::lang::Locale & aLocale) throw (css::uno::RuntimeException, css::lang::IndexOutOfBoundsException);
    virtual void SAL_CALL displayPropertyPage(::sal_Int32 nSmartTagIndex, const css::lang::Locale & aLocale) throw (css::uno::RuntimeException, css::lang::IndexOutOfBoundsException);

private:
    MOSTLSmartTagRecognizer(MOSTLSmartTagRecognizer &); // not defined
    void operator =(MOSTLSmartTagRecognizer &); // not defined

    virtual ~MOSTLSmartTagRecognizer();

    void LoadSmartTagFromFile( const rtl::OUString aPath, const rtl::OUString newfilename );
    void xPathError( xmlXPathContextPtr xpathCtx, xmlDocPtr doc );
    rtl::OUString getXmlValue( xmlXPathContextPtr xpathCtx, char *xpath );

    css::uno::Reference< css::uno::XComponentContext > m_xContext;
    sal_Int32 m_SmartTagCount;

    std::vector< MOSTLRecognizerData > m_aRecognizerData;
    rtl::OUString m_aComponentName;
};

MOSTLSmartTagRecognizer::MOSTLSmartTagRecognizer(css::uno::Reference< css::uno::XComponentContext > const & context) :
    m_xContext(context), m_SmartTagCount(0)
{}

MOSTLSmartTagRecognizer::~MOSTLSmartTagRecognizer()
{
}

// com.sun.star.uno.XServiceInfo:
::rtl::OUString SAL_CALL MOSTLSmartTagRecognizer::getImplementationName() throw (css::uno::RuntimeException)
{
    return comp_MOSTLSmartTagRecognizer::_getImplementationName();
}

::sal_Bool SAL_CALL MOSTLSmartTagRecognizer::supportsService(::rtl::OUString const & serviceName) throw (css::uno::RuntimeException)
{
    css::uno::Sequence< ::rtl::OUString > serviceNames = comp_MOSTLSmartTagRecognizer::_getSupportedServiceNames();
    for (::sal_Int32 i = 0; i < serviceNames.getLength(); ++i) {
        if (serviceNames[i] == serviceName)
            return sal_True;
    }
    return sal_False;
}

css::uno::Sequence< ::rtl::OUString > SAL_CALL MOSTLSmartTagRecognizer::getSupportedServiceNames() throw (css::uno::RuntimeException)
{
    return comp_MOSTLSmartTagRecognizer::_getSupportedServiceNames();
}

// ::com::sun::star::lang::XInitialization:
void SAL_CALL MOSTLSmartTagRecognizer::initialize(const css::uno::Sequence< ::css::uno::Any > & /*aArguments*/) throw (css::uno::RuntimeException, css::uno::Exception)
{
    std::vector<rtl::OUString> aFileList;
    rtl::OUString aDirectory;

    for ( sal_uInt32 j = 0; j < 2; ++j )
    {
        if ( 0 == j )
            ::utl::Bootstrap::locateSharedData( aDirectory );
        else
            ::utl::Bootstrap::locateUserData( aDirectory );

        lcl_GetFileList( aDirectory, aFileList );

        for ( sal_uInt32 i = 0; i < aFileList.size(); ++i )
            LoadSmartTagFromFile( aDirectory, aFileList[i] );

        aFileList.clear();
    }
}

// ::com::sun::star::smarttags::XSmartTagRecognizer:
::sal_Int32 SAL_CALL MOSTLSmartTagRecognizer::getSmartTagCount() throw (css::uno::RuntimeException)
{
    return m_SmartTagCount;
}

::rtl::OUString SAL_CALL MOSTLSmartTagRecognizer::getName(const css::lang::Locale & /*aLocale*/) throw (css::uno::RuntimeException)
{
    return m_aComponentName;
}

::rtl::OUString SAL_CALL MOSTLSmartTagRecognizer::getDescription(const css::lang::Locale & /*aLocale*/) throw (css::uno::RuntimeException)
{
    return rtl::OUString();
}

::rtl::OUString SAL_CALL MOSTLSmartTagRecognizer::getSmartTagName(::sal_Int32 nSmartTagIndex) throw (css::uno::RuntimeException, css::lang::IndexOutOfBoundsException)
{
  return m_aRecognizerData[nSmartTagIndex].getSmartTagType();
}

::rtl::OUString SAL_CALL MOSTLSmartTagRecognizer::getSmartTagDownloadURL(::sal_Int32 /*nSmartTagIndex*/) throw (css::uno::RuntimeException, css::lang::IndexOutOfBoundsException)
{
    return rtl::OUString();
}

void SAL_CALL MOSTLSmartTagRecognizer::recognize(const ::rtl::OUString & aText, ::sal_Int32 nStart, ::sal_Int32 nLength, css::smarttags::SmartTagRecognizerMode /*eDataType*/, const css::lang::Locale & aLocale, const css::uno::Reference< css::text::XTextMarkup > & xTextMarkup, const ::rtl::OUString & /*aApplicationName*/, const css::uno::Reference< css::frame::XController > & /*xController*/, const css::uno::Reference< css::i18n::XBreakIterator > & xTokenizer) throw (css::uno::RuntimeException)
{
    const sal_Int32 nEndPos = nStart + nLength;

    css::i18n::Boundary aWordBounds = xTokenizer->getWordBoundary( aText, nStart, aLocale, css::i18n::WordType::DICTIONARY_WORD, true );

    while ( aWordBounds.startPos < aWordBounds.endPos && aWordBounds.endPos <= nEndPos )
    {
        const rtl::OUString aWord = aText.copy( aWordBounds.startPos, aWordBounds.endPos - aWordBounds.startPos );

        /* search in termlist and regexps */
        for ( int i = 0; i < m_SmartTagCount; ++i )
        {
            if ( m_aRecognizerData[i].checkTermList( aWord ) ||
                 m_aRecognizerData[i].checkRegExps( aWord ) )
            {
                const rtl::OUString aSmartTagType = getSmartTagName( i );
                const sal_Int32 nSmartTagStart = aWordBounds.startPos;
                const sal_Int32 nSmartTagLength = aWordBounds.endPos - aWordBounds.startPos;
                css::uno::Reference< css::container::XStringKeyMap > xProps = xTextMarkup->getMarkupInfoContainer();
                xTextMarkup->commitTextMarkup( css::text::TextMarkupType::SMARTTAG,
                                                   aSmartTagType, nSmartTagStart, nSmartTagLength, xProps );
            }
        }

        aWordBounds = xTokenizer->nextWord( aText, aWordBounds.startPos, aLocale, css::i18n::WordType::DICTIONARY_WORD );
    }
}

::sal_Bool SAL_CALL MOSTLSmartTagRecognizer::hasPropertyPage(::sal_Int32 /*nSmartTagIndex*/, const css::lang::Locale & /*aLocale*/) throw (css::uno::RuntimeException, css::lang::IndexOutOfBoundsException)
{
    return sal_False;
}

void SAL_CALL MOSTLSmartTagRecognizer::displayPropertyPage(::sal_Int32 /*nSmartTagIndex*/, const css::lang::Locale & /*aLocale*/) throw (css::uno::RuntimeException, css::lang::IndexOutOfBoundsException)
{
}


} // closing anonymous implementation namespace



// component helper namespace
namespace comp_MOSTLSmartTagRecognizer {

::rtl::OUString SAL_CALL _getImplementationName() {
    return ::rtl::OUString(RTL_CONSTASCII_USTRINGPARAM(
        "MOSTLSmartTagRecognizer"));
}

css::uno::Sequence< ::rtl::OUString > SAL_CALL _getSupportedServiceNames()
{
    css::uno::Sequence< ::rtl::OUString > s(1);
    s[0] = ::rtl::OUString(RTL_CONSTASCII_USTRINGPARAM(
        "com.sun.star.smarttags.SmartTagRecognizer"));
    return s;
}

css::uno::Reference< css::uno::XInterface > SAL_CALL _create(
    const css::uno::Reference< css::uno::XComponentContext > & context)
        SAL_THROW((css::uno::Exception))
{
    return static_cast< ::cppu::OWeakObject * >(new MOSTLSmartTagRecognizer(context));
}

} // closing component helper namespace

static ::cppu::ImplementationEntry const entries[] = {
    { &comp_MOSTLSmartTagRecognizer::_create,
      &comp_MOSTLSmartTagRecognizer::_getImplementationName,
      &comp_MOSTLSmartTagRecognizer::_getSupportedServiceNames,
      &::cppu::createSingleComponentFactory, 0, 0 },
    { 0, 0, 0, 0, 0, 0 }
};

extern "C" void SAL_CALL component_getImplementationEnvironment(
    const char ** envTypeName, uno_Environment **)
{
    *envTypeName = CPPU_CURRENT_LANGUAGE_BINDING_NAME;
}

extern "C" void * SAL_CALL component_getFactory(
    const char * implName, void * serviceManager, void * registryKey)
{
    return ::cppu::component_getFactoryHelper(
        implName, serviceManager, registryKey, entries);
}

extern "C" sal_Bool SAL_CALL component_writeInfo(
    void * serviceManager, void * registryKey)
{
    return ::cppu::component_writeInfoHelper(serviceManager, registryKey, entries);
}


//
//
//
void MOSTLSmartTagRecognizer::xPathError(xmlXPathContextPtr xpathCtx, xmlDocPtr doc)
{
     fprintf(stderr,"Error: unable to evaluate xpath expression.\n");
     xmlXPathFreeContext(xpathCtx);
     xmlFreeDoc(doc);
}

rtl::OUString MOSTLSmartTagRecognizer::getXmlValue(xmlXPathContextPtr xpathCtx, char *xpath)
{
    xmlXPathObjectPtr xpathObj = xmlXPathEvalExpression((xmlChar*) xpath, xpathCtx);
    if(xpathObj == NULL)
        return C2U("");

    xmlNodeSetPtr nodes = xpathObj->nodesetval;
    int size = (nodes) ? nodes->nodeNr : 0;

    rtl::OUString aValue = C2U("");
    if (size>0)
    {
        const sal_Char* pTmp= (const sal_Char*) nodes->nodeTab[0]->children->content;
        aValue = C2U(pTmp);
    }

    xmlXPathFreeObject(xpathObj);

  return aValue;
}

void MOSTLSmartTagRecognizer::LoadSmartTagFromFile( const rtl::OUString aPath, const rtl::OUString newfilename )
{
    const rtl::OString ostring( ::rtl::OUStringToOString (newfilename, RTL_TEXTENCODING_UTF8) );

    xmlDocPtr doc = xmlParseFile(ostring.getStr());

    if (doc == NULL )
    {
        fprintf(stderr,"Document not parsed successfully. \n");
        return;
    }

    /* Create xpath evaluation context */
    xmlXPathContextPtr xpathCtx = xmlXPathNewContext(doc);
    if(xpathCtx == NULL)
    {
      fprintf(stderr,"Error: unable to create new XPath context\n");
      xmlFreeDoc(doc);
      return;
    }

    /* register namespace */
    if(xmlXPathRegisterNs(xpathCtx, (xmlChar*) "FL", (xmlChar*) "http://schemas.microsoft.com/office/smarttags/2003/mostl") != 0)
    {
        fprintf(stderr,"Error: unable to register NS with prefix=\"%s\" and href=\"%s\"\n", "", "");
        xmlFreeDoc(doc);
        return;
    }

    /* get name */
    m_aComponentName = C2U("Smart Tag List Recognizer");/*getXmlValue(xpathCtx, "/FL:smarttaglist/FL:name");*/

    /* get smarttag count */
    xmlXPathObjectPtr xpathObj = xmlXPathEvalExpression((xmlChar*) "/FL:smarttaglist/FL:smarttag", xpathCtx);
    if(xpathObj == NULL)
    {
        xPathError(xpathCtx, doc);
        return;
    }

    xmlNodeSetPtr nodes = xpathObj->nodesetval;
    int nNumberOfSmartTagTypes = (nodes) ? nodes->nodeNr : 0;

    m_SmartTagCount += nNumberOfSmartTagTypes;

    xmlXPathFreeObject(xpathObj);

    if ( !nNumberOfSmartTagTypes )
        return;

    // iterate over all smarttags in found in xml file
    for( int j = 1; j <= nNumberOfSmartTagTypes; ++j )
    {
        MOSTLRecognizerData aData;

        /* set caption */
        sal_Char xpath[128];
        snprintf(xpath, sizeof(xpath), "/FL:smarttaglist/FL:smarttag[%d]/FL:caption", j);
        aData.setCaption(getXmlValue(xpathCtx, xpath));

        /* get recognizer name */
        snprintf(xpath, sizeof(xpath), "/FL:smarttaglist/FL:smarttag[%d]/@type", j);
        aData.setSmartTagType(getXmlValue(xpathCtx, xpath));

        //
        // get term list
        //
        snprintf(xpath, sizeof(xpath), "/FL:smarttaglist/FL:smarttag[%d]/FL:terms/FL:termlist", j);
        xpathObj = xmlXPathEvalExpression((xmlChar*) xpath, xpathCtx);
        if(xpathObj == NULL)
        {
            xPathError(xpathCtx, doc);
            return;
        }

        nodes = xpathObj->nodesetval;
        int size = (nodes) ? nodes->nodeNr : 0;

        rtl::OUString aTList;
        for( int i = 0; i < size; ++i )
        {
            const sal_Char* pTmp= (const sal_Char*) nodes->nodeTab[i]->children->content;
            rtl::OUString aTmp(pTmp, strlen(pTmp), RTL_TEXTENCODING_UTF8);

            sal_Int32 nStart = 0;
            sal_Int32 nPos = -1;
            do
            {
                nPos = aTmp.indexOf( ',', nStart );
                sal_Int32 nEndPos = -1 != nPos ? nPos : aTmp.getLength();
                rtl::OUString aTerm = aTmp.copy( nStart, nEndPos - nStart );
                aTerm = aTerm.trim();
                aData.appendTermList(aTerm);

                nStart = nPos + 1;
            }
            while ( -1 != nPos );
        }

        xmlXPathFreeObject(xpathObj);

        //
        // get terms file list
        //
        snprintf(xpath, sizeof(xpath), "/FL:smarttaglist/FL:smarttag[%d]/FL:terms/FL:termfile/FL:filename", j);
        xpathObj = xmlXPathEvalExpression((xmlChar*) xpath, xpathCtx);
        if(xpathObj == NULL)
        {
            xPathError(xpathCtx, doc);
            return;
        }

        nodes = xpathObj->nodesetval;
        size = (nodes) ? nodes->nodeNr : 0;

        for( int i = 0; i < size; ++i )
        {
            const sal_Char* pTmp = (sal_Char*)nodes->nodeTab[i]->children->content;

            INetURLObject aObject( aPath );
            aObject.insertName( C2U("smarttags") );
            aObject.insertName( rtl::OStringToOUString( pTmp, RTL_TEXTENCODING_UTF8 ), false, INetURLObject::LAST_SEGMENT, true, INetURLObject::ENCODE_ALL );
            const rtl::OUString aMainURL = aObject.GetMainURL(INetURLObject::NO_DECODE);

            SvFileStream aStream( aMainURL, STREAM_READ );
            String aEntry;

            while ( aStream.ReadByteStringLine( aEntry, RTL_TEXTENCODING_UTF8 ) )
            {
                aData.appendTermList( aEntry );
            }
        }

        xmlXPathFreeObject(xpathObj);

        //
        // get regex
        //
        snprintf(xpath, sizeof(xpath), "/FL:smarttaglist/FL:smarttag[%d]/FL:re/FL:exp", j);
        xpathObj = xmlXPathEvalExpression((xmlChar*) xpath, xpathCtx);
        if(xpathObj == NULL)
        {
            xPathError(xpathCtx, doc);
            return;
        }

        nodes = xpathObj->nodesetval;
        size = (nodes) ? nodes->nodeNr : 0;

        for( int i = 0; i < size; ++i )
        {
            _xmlNode* children = nodes->nodeTab[i]->children;
            // search regexp string:
            do
            {
                const sal_Char* pName = (const sal_Char*) children->name;
                if ( 0 == strcmp( pName, "text" ) )
                    break;

                children = children->next;
            }
            while ( 0 != children );

            if ( children )
            {
                const sal_Char* pTmp= (const sal_Char*) children->content;
                rtl::OUString aTmp(pTmp, strlen(pTmp), RTL_TEXTENCODING_UTF8);
                aData.appendRegexList(aTmp);
            }
        }

        xmlXPathFreeObject(xpathObj);

        m_aRecognizerData.push_back( aData );
    }

    xmlXPathFreeContext(xpathCtx);
    xmlFreeDoc(doc);
}
